'''
Module containing high-level evogression functionality to fit and summarize regressions.
'''
from typing import List, Dict, Union
import random
from math import log
from collections import defaultdict
import pandas
from pandas import DataFrame
from .evolution import Evolution



def evolution_group(target_param: str, data: Union[List[Dict[str, float]], DataFrame], num_creatures: int=10000, num_cycles: int=10,
                    max_layers: int=3, group_size: int=4, max_cpu: int=0, optimize=True, **kwargs) -> List[Evolution]:
    '''
    Generate a list of fully initialized Evolution objects.
    Any Evolution kwargs may be provided.
    '''
    return [Evolution(target_param, data, num_creatures=num_creatures, num_cycles=num_cycles, max_layers=max_layers,  max_cpu=max_cpu, optimize=optimize, **kwargs) for _ in range(group_size)]


def output_group_regression_funcs(group: list):
    '''
    Take list of Evolution objects and output their
    best regressions to a "regression_modules" subdir.
    '''
    for cr_ev in group:
        cr_ev.output_best_regression(directory='regression_modules', add_error_value=True)


def parameter_usage(group: list) -> Dict[str, int]:
    '''
    Combine each Evolution's .parameter_usefulness_count dicts to see which attributes are important.
    '''
    combined_parameter_usefulness = defaultdict(int)
    for cr_ev in group:
        for param, count in cr_ev.parameter_usefulness_count.items():
            combined_parameter_usefulness[param] += count
    return combined_parameter_usefulness


def output_usage(group: list, filename: str='ParameterUsage.xlsx'):
    usage = parameter_usage(group)
    usage = pandas.DataFrame.from_dict(usage, orient='index')
    usage.reset_index(inplace=True)
    usage.columns = ['PARAMETER', 'USAGE']
    usage.sort_values('USAGE', ascending=False, inplace=True)
    usage.to_excel(filename if '.' in filename else filename + '.xlsx', index=False)


def generate_parameter_usage_file(data, column: str, num_models: int=100, num_creatures: int=5000, num_cycles: int=3, num_cpu: int=1, filename: str='ParameterUsage.xlsx'):
    '''
    Generate many models using subsets of possible input data columns so as to
    not overweight usage of the few best columns.
    Output the parameter usage/predictability of the data columns to Excel.

    CURRENTLY only accepts a DataFrame for data arg.
    '''
    columns = list(data.columns)
    if len(columns) <= 4:
        num_col_per_sample = len(columns)
    else:
        num_col_per_sample = int(7 * log(len(columns)) - 8) + 1

    models = []
    for _ in range(num_models):
        col_subset = random.sample(columns, num_col_per_sample)
        if column not in col_subset:  # need prediction column in the data
            col_subset.append(column)
        data_subset = data[col_subset]
        model = Evolution(column, data_subset, num_creatures=num_creatures, num_cycles=num_cycles, num_cpu=num_cpu, optimize=False)
        model.clear_data()  # clear out references to data to shrink model object and limit memory growth
        models.append(model)
    output_usage(models, filename)


def parameter_pruned_evolution_group(target_param: str, data: list, max_parameters: int=10, num_creatures: int=10000, num_cycles: int=10, group_size: int=4) -> List[Evolution]:
    '''
    Generate successive groups of Evolution objects and prune least-used
    parameters from the input data each round until only the most useful parameters remain.
    Finally, run a full-blown evolution cycles with the remaining parameters
    for the saved regression modules.

    USE LARGE >>> num_cycles <<< TO INTEGRATE MORE STATISTICALLY VALID PARAMETER USAGE NUMBERS BEFORE REMOVING PARAMETERS
    '''
    if num_creatures < 1000 or num_cycles < 10:
        print('\n> ERROR!  parameter_pruned_evolution_group() can be unstable (infinite loop)\n  if num_creatures and/or num_cycles args are too small.')
        print('  Please use a minimum num_creatures of 1000 and a minimum num_cycles of 10.\n  Higher values num_cycles are encouraged!\n')
        return

    def num_param_to_eliminate(num_extra_param: int) -> int:
        '''Determine how many worst-performing parameters to elimintate in a given round'''
        if num_extra_param > 50:
            return num_extra_param // 2
        elif num_extra_param > 10:
            return num_extra_param // 3
        elif num_extra_param > 3:
            return 2
        elif num_extra_param > 0:
            return 1
        else:
            return 0

    num_parameters = len(data[0].keys()) - 1
    while num_parameters > max_parameters:
        group = evolution_group(target_param, data, int(num_creatures // 1.6), int(num_cycles // 1.6), group_size=group_size, optimize=False)

        current_parameter_usage = [(param, count) for param, count in parameter_usage(group).items()]
        random.shuffle(current_parameter_usage)  # so below filter ignores previous order for equally-ranked parameters
        ranked_parameters = sorted(current_parameter_usage, key=lambda tup: tup[1])
        print(ranked_parameters)

        dead_params = [t[0] for t in ranked_parameters[:num_param_to_eliminate(num_parameters - max_parameters)]]
        for data_point in data:
            for param in dead_params:
                del data_point[param]
        num_parameters = len(data[0].keys()) - 1

    final_group = evolution_group(target_param, data, num_creatures, num_cycles, group_size=group_size)
    print('parameter_pruned_evolution_group complete.  Final Parameter usage counts below:')
    for param, count in parameter_usage(final_group).items():
        print(f'  {count}: {param}')
    return final_group


def random_population(target_param: str, data: list, num_creatures: int=10000, num_cycles: int=10, group_size: int=4, **kwargs) -> List[Evolution]:
    '''
    Generate a list of Evolution objects (same as evolution_group) but use randomly sampled data subsets for training.
    The goal is to generate a "Random Population" in a similar manner as a Random Forest concept.
    '''
    data_subset_size = int(len(data) * 1.5 // group_size)
    evolutions = []
    for _ in range(group_size):
        data_subset = random.choices(data, k=data_subset_size)
        evolutions.append(Evolution(target_param, data_subset, num_creatures=num_creatures, num_cycles=num_cycles, clear_creatures=True, **kwargs))
    return evolutions


class Population():
    def __init__(self, target_param: str, data: list, num_creatures=300, num_cycles: int=3, group_size: int=4, split_parameter=None, category_or_continuous='category', bin_size=None, num_cpu=1, **kwargs):
        self.target_parameter = target_param

        if type(data) == DataFrame:
            data = data.to_dict('records')

        self.split_parameter = split_parameter
        self.category_or_continuous = category_or_continuous
        if split_parameter and category_or_continuous == 'category':
            categories = set(d[split_parameter] for d in data)
            data_sets = {cat: [{k: v for k, v in d.items() if k != split_parameter} for d in data if d[split_parameter] == cat] for cat in categories}
            self.evo_sets = {}
            for cat, data_subset in data_sets.items():
                self.evo_sets[cat] = [Evolution(target_param, data_subset, num_creatures=num_creatures, num_cycles=num_cycles, clear_creatures=True, num_cpu=num_cpu, **kwargs) for _ in range(group_size)]

        elif split_parameter and category_or_continuous == 'continuous':
            # Use bin_size arg (or generate if not provided) to determine how to split out data into different bins of the split_parameter.
            split_values = sorted(d[split_parameter] for d in data)
            if not bin_size:
                bin_size = (max(split_values) - min(split_values)) / 5

            bins = []
            binned_data = {}
            lower = min(split_values)
            upper = lower + bin_size
            while lower < max(split_values):
                bins.append((lower, upper))
                cutoff_lower = lower - bin_size * 0.5
                cutoff_upper = upper + bin_size * 0.5
                binned_data[(lower, upper)] = [d for d in data if cutoff_lower <= d[split_parameter] < cutoff_upper]
                lower, upper = upper, upper + bin_size
            self.bins = bins

            self.evo_sets = {}
            for bin in bins:
                self.evo_sets[bin] = evolution_group(target_param, binned_data[bin], num_creatures=num_creatures, num_cycles=num_cycles, group_size=group_size, **kwargs)



    def predict(self, data: Union[Dict[str, float], List[Dict[str, float]], DataFrame], prediction_key: str=''):
        '''
        Generate predictions with same interface as BaseEvolution.predict.
        '''
        if prediction_key == '':
            prediction_key = f'{self.target_parameter}_PREDICTED'

        is_dataframe = True if type(data) == DataFrame else False
        if is_dataframe:
            data = data.to_dict('records')  # will get processed as a list

        if self.category_or_continuous == 'category':
            if isinstance(data, list):  # dataframes also get processed here
                for d in data:
                    predictions = [evo.predict(d, 'pred')['pred'] for evo in self.evo_sets[d[self.split_parameter]]]
                    d[prediction_key] = sum(predictions) / len(predictions)
                return DataFrame(data) if is_dataframe else data
            elif isinstance(data, dict):
                predictions = [evo.predict(data, 'pred')['pred'] for evo in self.evo_sets[data[self.split_parameter]]]
                data[prediction_key] = sum(predictions) / len(predictions)
                return data
            else:
                print('Error!  "data" arg provided to .predict() must be a dict or list of dicts or DataFrame.')


        elif self.category_or_continuous == 'continuous':
            if isinstance(data, list):  # dataframes also get processed here
                data_points = data
                is_dict = False
            elif isinstance(data, dict):
                data_points = [data]
                is_dict = True

            bin_dist = lambda val, bin: abs(val - (bin[1]+bin[0]) / 2)
            for data_point in data_points:
                bin_distances = sorted([(bin, bin_dist(data_point[self.split_parameter], bin)) for bin in self.bins], key=lambda tup: tup[1])
                min_dist = bin_distances[0][1] if bin_distances[0][1] > 0 else bin_distances[1][1]
                bin_distances = [(t[0], t[1] + min_dist) for t in bin_distances]  # decrease distance sensitivity a hair (still include other weight is closest is almost exactly the value)
                bin_distances = bin_distances[:-2 * int(len(bin_distances) / 3)]  # remove furthest 2/3 of bins from consideration
                total_dists = sum(t[1] for t in bin_distances)

                # Now generate weighted-average prediction by giving more weight to closer bins and less to further ones
                total_prediction = 0.0
                for bin, dist in bin_distances:
                    predictions = sorted(evo.predict(data_point, 'pred')['pred'] for evo in self.evo_sets[bin])
                    bin_pred = sum(predictions[1:-1]) / (len(predictions) - 2)  # kick out max/min (outlyer) predictions
                    total_prediction += bin_pred * (dist / total_dists)

                data_point[prediction_key] = total_prediction

            if is_dict:
                return data_points[0]
            elif is_dataframe:
                return DataFrame(data_points)
            else:
                return data_points



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import linear_data
import matplotlib.pyplot as plt


class TestLinearRegressionEvolution(unittest.TestCase):
    def test_best_creature_evolution(self):
        evolution = evogression.Evolution('y', linear_data, num_creatures=10000, num_cycles=3)

        predictions = [{'x': i / 2} for i in range(6, 25)]
        predictions = evolution.predict(predictions)
        calculation_x_values = [point['x'] for point in predictions]
        calculated_y_values = [point['y_PREDICTED'] for point in predictions]

        plt.scatter([d['x'] for d in linear_data], [d['y'] for d in linear_data])
        plt.plot(calculation_x_values, calculated_y_values, 'g--')

        plt.xlabel('x')
        plt.ylabel('y')
        plt.title('Linear Regression - Evolution Test')
        plt.show()


if __name__ == '__main__':
    unittest.main()



'''
Module containing evolution algorithms for regression.
'''
from typing import List, Dict, Union
from collections import defaultdict
import os
import pickle
from pandas import DataFrame
from . import data
from . import rust_evogression


class Evolution():
    '''
    EVOLUTION ALOGORITHM

    Evolves creatures by killing off the worst performers in
    each cycle and then randomly generating many new creatures.

    Input data must have all numeric values (or None).
    '''
    def __init__(self,
                 target_parameter: str,
                 all_data: Union[List[Dict[str, float]], DataFrame],
                 num_creatures: int=10000,
                 num_cycles: int=10,
                 max_layers: int=3,
                 max_cpu: int=max(os.cpu_count()-1, 1),  # by default use all but one core
                 optimize: bool=True,
                 **kwargs) -> None:

        self.target_parameter = target_parameter

        if isinstance(all_data, DataFrame):
            all_data = all_data.to_dict('records')

        self.param_medians = data.calc_param_medians(all_data, target_parameter)
        self.all_data = data.fill_none_with_median(all_data, target_parameter, self.param_medians)
        self.full_parameter_example = self.all_data[0]

        data.data_checks(self.all_data)

        self.num_creatures = int(num_creatures)
        self.num_cycles = int(num_cycles)
        self.max_layers = int(max_layers)

        os.environ['RAYON_NUM_THREADS'] = str(max_cpu)
        self.model = rust_evogression.run_evolution(target_parameter, self.all_data, num_creatures, num_cycles, max_layers, optimize)

        self.parameter_usefulness_count: dict = defaultdict(int)
        for creature in self.model.best_creatures:
            for param in creature.used_parameters():
                self.parameter_usefulness_count[param] += 1


    @property
    def best_creature(self):
        return self.model.best_creature

    @property
    def best_error(self):
        return self.model.best_error()


    def output_best_regression(self, output_filename='regression_function', directory: str='.', add_error_value=False) -> None:
        '''
        Save this the regression equation/function this evolution has found
        to be the best into a new Python module so that the function itself
        can be imported and used in other code.
        '''
        name_ext = f'___{round(self.model.best_error(), 4)}' if add_error_value else ''

        if directory != '.' and not os.path.exists(directory):
            os.mkdir(directory)

        if output_filename[-3:] == '.py':  # adding .py later; removing to easily check for periods
            output_filename = output_filename[:-3]

        output_filename = output_filename.replace('.', '_') # period in filename not valid
        output_filename = os.path.join(directory, output_filename + name_ext.replace('.', '_') + '.py')

        output_str = self.model.python_regression_module_string()
        with open(output_filename, 'w') as f:
            f.write(output_str)
        print('Evogression model saved as a Python module.')


    def save(self, filename='evolution_model') -> None:
        '''
        Save this Evolution Python object/model in a pickle file.
        Also removes non-essential data to reduce file size.
        '''
        # Clearing or shrinking unneeded attributes provides a smaller file size.
        self.all_data = None

        with open(filename if '.' in filename else filename + '.pkl', 'wb') as f:
            pickle.dump(self, f)


    @classmethod
    def load(cls, filename):
        '''
        Load an Evolution object from a saved, pickle file.
        '''
        try:
            with open(filename, 'rb') as f:
                return pickle.load(f)
        except:
            with open(filename + '.pkl', 'rb') as f:
                return pickle.load(f)


    def predict(self, data: Union[Dict[str, float], List[Dict[str, float]], DataFrame], prediction_key: str='', standardized_data: bool=False):
        '''
        Add best_creature predictions to data arg as f'{target}_PREDICTED' new key.
        Return unstandardized dict or list of dicts or DataFrame depending on provided arg.
        '''
        param_medians = self.param_medians  # local variable for speed
        param_example = self.full_parameter_example  # local variable for speed

        if prediction_key == '':
            prediction_key = f'{self.target_parameter}_PREDICTED'

        is_dataframe = True if isinstance(data, DataFrame) else False
        if is_dataframe:
            data = data.to_dict('records')  # will get processed as list

        if isinstance(data, list):  # DataFrames also get processed here (previously converted to a list)
            for d in data:
                for param, val in d.items():
                    if not val:  # make any None values the previously calculated median from the training data
                        d[param] = param_medians.get(param, 0.0)
                # Now remove any keys in data not in training data as will not be in regression and can cause issues.  Also don't want string values.
                clean = {k: v for k, v in d.items() if k in param_example and not isinstance(v, str)}
                d[prediction_key] = self.model.predict_point(clean)

            if is_dataframe:
                data = DataFrame(data)
            return data

        elif isinstance(data, dict):
            # make any None values the previously calculated median from the training data
            for param, val in data.items():
                if not val:
                    data[param] = param_medians.get(param, 0.0)
            # Now remove any keys in data not in training data as will not be in regression and can cause issues.  Also don't want string values.
            data = {k: v for k, v in data.items() if k in param_example and not isinstance(v, str)}
            data[prediction_key] = self.model.predict_point(data)
            return data

        else:
            print('Error!  "data" arg provided to .predict() must be a dict or list of dicts or DataFrame.')



import math
import statistics


def data_checks(data_to_check: list[dict]) -> None:
    '''
    Check cleaned input data for potential issues.
    At the point when this is called, there should be no issues with
    the data to be used; data-cleaning methods are called earlier.

    If this method prints errors, we need to write more data-cleaning capabilities to handle those cases!
    '''
    acceptable_types = {'float', 'int', 'float64', 'int64'}
    issues = []
    def check_data(data, data_name):
        for i, d in enumerate(data):
            for key, val in d.items():
                # val > & < checks are way of checking for nan without needing to require numpy import
                if type(val).__name__ not in acceptable_types or not (val >= 0 or val <= 0):
                    issues.append((data_name, i, key, val))

    check_data(data_to_check, 'all_data')
    for issue in issues:
        data_name, i, key, val = issue
        print(f'\nERROR!  NAN values detected in {data_name}!')
        print(f'Index: {i}  key: {key}  value: {val}  type: {type(val).__name__}')


def calc_param_medians(data: list[dict], target: str) -> dict:
     # Remove any data points that have None for the target/result parameter
    data = [d for d in data if d[target] is not None and not math.isnan(d[target])]

    is_nan = math.isnan  # local for speed
    param_medians = {}
    for param in data[0].keys():
        if param != target:
            values = [val for d in data if (val := d[param]) is not None and not is_nan(val)]
            param_medians[param] = statistics.median(values) if values else 0.0  # if values is empty, set all to zero
    return param_medians


def fill_none_with_median(data: list[dict], target: str, param_medians: dict) -> list[dict]:
        '''
        Find median value of each input parameter and
        then replace any None values with this median.
        '''
        # Remove any data points that have None for the target/result parameter
        data = [d for d in data if d[target] is not None and not math.isnan(d[target])]

        # param_medians = {}  # used for default parameter values if None is provided in .predict
        is_nan = math.isnan  # local for speed
        parameters_adjusted = set()
        for param in (key for key in data[0].keys() if key != target):
            median = param_medians[param]
            for d in data:
                if (val := d[param]) is None or is_nan(val):
                    parameters_adjusted.add(param)
                    d[param] = median

        if len(parameters_adjusted) >= 1:
            print('Data None values filled with median for the following parameters:')
            for param in sorted(parameters_adjusted):
                print(f'  {param}')

        return data


class InputDataFormatError(Exception):
    def __init__(self, *args):
        self.message = args[0] if args else None

    def __str__(self):
        return '\n' + str(self.message) if self.message else '\n'



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import linear_data
import matplotlib.pyplot as plt


class TestGroupComposite(unittest.TestCase):
    def test_evolution_group(self):
        group = evogression.evolution_group('y', linear_data, num_creatures=25, group_size=100,
            num_cycles=5, progressbar=False)

        calculation_x_values = [i / 2 for i in range(6, 27)]
        for evo in group:
            calculated_y_values = [evo.predict({'x': x}, 'pred')['pred'] for x in calculation_x_values]
            plt.plot(calculation_x_values, calculated_y_values, alpha=0.1)

        calculated_y_values = [sum(evo.predict({'x': x}, 'pred')['pred'] for evo in group) / len(group) for x in calculation_x_values]
        plt.plot(calculation_x_values, calculated_y_values,  'g--', alpha=1.0)

        plt.scatter([d['x'] for d in linear_data], [d['y'] for d in linear_data])

        plt.xlabel('x')
        plt.ylabel('y')
        plt.title('Linear Regression - Group Composite Prediction Test')
        plt.show()

        evogression.output_usage(group)



if __name__ == '__main__':
    unittest.main(buffer=True)



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import surface_3d_data
import pandas



class TestDataFrame(unittest.TestCase):
    def test_load_and_predict(self):
        '''
        Test ability of Evolution class to handle pandas DataFrames.
        '''
        df = pandas.DataFrame(surface_3d_data)
        evolution = evogression.Evolution('z', df, num_creatures=500, num_cycles=5, use_multip=False)
        predicted = evolution.predict(df)
        self.assertTrue(type(predicted) == pandas.DataFrame)
        self.assertTrue(len(predicted.columns) == 4)
        self.assertTrue('z_PREDICTED' in predicted.columns)

        single_prediction = evolution.predict({'x': 0, 'y': 0}, 'z_test')
        self.assertTrue('z_test' in single_prediction)
        self.assertTrue(len(single_prediction) == 3)





if __name__ == '__main__':
    unittest.main(buffer=True)



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import linear_data
from test_data_many_dimensions import data as many_d_data


class TestGroupFunctions(unittest.TestCase):
    def test_parameter_usage_count(self):
        group = evogression.groups.evolution_group('Target', many_d_data, num_creatures=500, num_cycles=5, group_size=10, optimize=False)
        for ev in group:
            print(ev.parameter_usefulness_count)
            self.assertTrue(len(ev.parameter_usefulness_count) > 0)



if __name__ == '__main__':
    unittest.main(buffer=False)



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import linear_data
from test_data_many_dimensions import data as many_d_data
from pprint import pprint as pp



class TestPredictionMethods(unittest.TestCase):
    def test_predict(self):
        '''
        Test ability of Evolution class to add target variable predictions
        to data samples.
        '''
        try:
            evolution = evogression.Evolution('y', linear_data, num_creatures=1000, num_cycles=5, use_multip=False)
            print(evolution.predict({'x': 5.2}))
            print(evolution.predict([{'x': -3.8}, {'x': 2.9}, {'x': 12.4}]))
            test_passed = True
        except TypeError:
            test_passed = False
        self.assertTrue(test_passed)

    def test_param_usage_counts(self):
        evolution = evogression.Evolution('Target', many_d_data, num_creatures=300, num_cycles=30, use_multip=False, optimize=False)
        pp(evolution.parameter_usefulness_count)
        d = dict(evolution.parameter_usefulness_count)
        self.assertTrue(len(d) > 1)
        self.assertTrue(sum(d.values()) > 3)

    # def test_save_and_load(self):
        # evolution = evogression.Evolution('y', linear_data, num_creatures=1000, num_cycles=5, use_multip=False)
        # prediction1 = evolution.predict({'x': 5.2})
        # evolution.save('linear_model')
        # loaded_model = evogression.Evolution.load('linear_model')
        # os.remove('linear_model.pkl')
        # prediction2 = loaded_model.predict({'x': 5.2})
        # self.assertTrue(prediction1 == prediction2)




if __name__ == '__main__':
    unittest.main(buffer=True)



from ._version import __version__

from . import evolution
from .evolution import Evolution
# from .creatures import EvogressionCreature
from . import groups
from .groups import evolution_group, parameter_usage, output_usage, random_population, Population



import unittest
import cProfile
import sys
sys.path.insert(1, '..')
import evogression
from test_data_many_dimensions import data as many_d_data



class TestManyDimensionParameterPruningRegression(unittest.TestCase):
    def test_best_creature_parabola_regression_evolution(self):
        pruned_evolution_group = evogression.groups.parameter_pruned_evolution_group('Target', many_d_data, max_parameters=15, num_creatures=1000, num_cycles=10, group_size=3)
        evogression.groups.output_group_regression_funcs(pruned_evolution_group)
        self.assertTrue(True)



if __name__ == '__main__':
    # have to do the below magic to make cProfile work with unittest
    suite = unittest.TestLoader().discover('.')
    def run_tests():
        unittest.TextTestRunner().run(suite)
    cProfile.run('unittest.main()', 'parameter_pruning_test.profile')



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import categorical_data, surface_3d_data
from pprint import pprint as pp
import matplotlib.pyplot as plt


class TestPopulationCateogry(unittest.TestCase):

    def test_population_category_2d(self):
        population = evogression.Population('y', categorical_data, split_parameter='cat', num_creatures=3000)
        y_test = [population.predict(d, 'pred')['pred'] for d in categorical_data]

        plt.scatter([pd['x'] for pd in categorical_data], [pd['y'] for pd in categorical_data], s=100)

        x_test_A = [i / 10 for i in range(0, 55)]
        y_test_A = [population.predict({'cat': 'A', 'x': x}, 'pred')['pred'] for x in x_test_A]

        x_test_B = [i / 10 for i in range(87, 135)]
        y_test_B = [population.predict({'cat': 'B', 'x': x}, 'pred')['pred'] for x in x_test_B]

        plt.scatter(x_test_A, y_test_A, s=10)
        plt.scatter(x_test_B, y_test_B, s=10)

        plt.xlabel('x')
        plt.ylabel('y')
        plt.text(1, 17, 'Category A')
        plt.text(10.5, 17, 'Category B')
        plt.title('Category Regression - Population Test')

        plt.show()


    def test_population_continuous_3d(self):
        population = evogression.Population('z', surface_3d_data, num_creatures=5000, num_cycles=7, group_size=5, split_parameter='y', category_or_continuous='continuous')
        data = population.predict(surface_3d_data, 'z_predicted')

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        x = [point_dict['x'] for point_dict in data]
        y = [point_dict['y'] for point_dict in data]
        z = [point_dict['z'] for point_dict in data]
        z_test = [point_dict['z_predicted'] for point_dict in data]

        ax.scatter3D(x, y, z)
        ax.scatter3D(x, y, z_test)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        plt.title('Surface Regression - Continuous Population Test')

        plt.show()



if __name__ == '__main__':
    unittest.main()



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import surface_3d_data
from pprint import pprint as pp
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d

import random
random.seed(10)  # for reproducing the same plot


class Test3DSurfaceRegression(unittest.TestCase):

    def test_best_creature_3d(self):
        evolutions = evogression.random_population('z', surface_3d_data, num_creatures=5000, num_cycles=5, group_size=15)
        z_test = [sum(e.predict(d, 'pred')['pred'] for e in evolutions) / len(evolutions) for d in surface_3d_data]

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        x = [point_dict['x'] for point_dict in surface_3d_data]
        y = [point_dict['y'] for point_dict in surface_3d_data]
        z = [point_dict['z'] for point_dict in surface_3d_data]

        ax.scatter3D(x, y, z)
        ax.scatter3D(x, y, z_test)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        plt.title('Surface Regression - Random Population Test')

        plt.show()



if __name__ == '__main__':
    unittest.main()



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import linear_data


class TestNoneFilling(unittest.TestCase):

    def test_none_fill(self):
        '''
        Test ability of CreatureEvolution class to handle None values in input data
        by filling them with the median of populated values.
        '''
        test_data = linear_data
        test_data[1] = {'x': 5, 'y': None}
        test_data[4] = {'x': None, 'y': 11.2}
        test_data[5] = {'x': float('nan'), 'y': 11.2}
        test_data[6] = {'x': None, 'y': float('nan')}
        try:
            evolution = evogression.Evolution('y', test_data, num_creatures=100, num_cycles=1, use_multip=False)
            test_passed = True
        except TypeError:
            test_passed = False
        self.assertTrue(test_passed)



if __name__ == '__main__':
    unittest.main(buffer=True)



import unittest
import cProfile
import sys
sys.path.insert(1, '..')
import evogression
from test_data import parabola_data

import matplotlib.pyplot as plt


class TestParabolaRegression(unittest.TestCase):

    def test_best_creature_parabola_regression_evolution(self):
        model = evogression.Evolution('y', parabola_data, num_creatures=10000, num_cycles=10, force_num_layers=0, standardize=True)
        model.output_best_regression(directory='regression_modules', add_error_value=True)

        x_values = list(range(-20, 21))
        y_values = [model.predict({'x': x})['y_PREDICTED'] for x in x_values]

        plt.scatter([d['x'] for d in parabola_data], [d['y'] for d in parabola_data])
        plt.plot(x_values, y_values, 'g--')

        plt.xlabel('x')
        plt.ylabel('y')
        plt.title('Parabola Regression - Evolution Test')
        plt.show()



if __name__ == '__main__':
    # have to do the below magic to make cProfile work with unittest
    suite = unittest.TestLoader().discover('.')
    def run_tests():
        unittest.TextTestRunner().run(suite)
    cProfile.run('unittest.main()', 'parabola_test.profile')



from setuptools import setup, find_packages
from setuptools_rust import Binding, RustExtension
from evogression import __version__


with open('README.md', 'r') as f:
    long_description = f.read()


setup(name='evogression',
      version=__version__,
      packages=find_packages(),
      license='MIT',
      author='Zach Bateman',
      description='Evolutionary Regression',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/zachbateman/evogression.git',
      download_url='https://github.com/zachbateman/evogression/archive/v_' + __version__ + '.tar.gz',
      keywords=['REGRESSION', 'MACHINE', 'LEARNING', 'EVOLUTION'],
      install_requires=['pandas'],
      classifiers=['Development Status :: 4 - Beta',
                    'License :: OSI Approved :: MIT License',
                    'Programming Language :: Python :: 3.9',
                    'Programming Language :: Python :: 3.10',
                    'Programming Language :: Python :: 3.11',
                    ],
    rust_extensions=[RustExtension("evogression.rust_evogression", binding=Binding.PyO3)]
)



import unittest
import sys
sys.path.insert(1, '..')
import evogression
from test_data import surface_3d_data
import matplotlib.pyplot as plt


class Test3DSurfaceRegression(unittest.TestCase):

    def test_best_creature_3d(self):
        evolution = evogression.Evolution('z', surface_3d_data, num_creatures=30000, num_cycles=10, num_cpu=3)
        z_test = [evolution.predict(d, 'pred')['pred'] for d in surface_3d_data]

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        x = [point_dict['x'] for point_dict in surface_3d_data]
        y = [point_dict['y'] for point_dict in surface_3d_data]
        z = [point_dict['z'] for point_dict in surface_3d_data]

        ax.scatter3D(x, y, z)
        ax.scatter3D(x, y, z_test)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        plt.title('Surface Regression - Evolution Test')

        plt.show()



if __name__ == '__main__':
    unittest.main()
